
object CPoint {

  // let's keep things simple -- only use a 
  // cartesian creators for color points

  def cartesian(x:Double,y:Double,c:Color):CPoint =
    new CartesianCPoint(x,y,c)
  


  private 
  class CartesianCPoint (xpos:Double, ypos:Double, c:Color) extends CPoint {

    // defining the delegate
    val del:Point = Point.cartesian(xpos,ypos)

    // easy delegations
    def xCoord ():Double = del.xCoord()
    def yCoord ():Double = del.yCoord()
    def distanceFromOrigin ():Double = del.distanceFromOrigin()
    def angleWithXAxis ():Double = del.angleWithXAxis()
    def distance (q:CPoint):Double = del.distance(q)
    def isOrigin ():Boolean = del.isOrigin()

    // delegations where we do something extra
    def isEqual (q:CPoint):Boolean = {
      del.isEqual(q) && c==q.color()
    }

    // delegations where we need to "rebuild" a CPoint via helper method:

    private def addColor(p:Point,c:Color):CPoint = 
      cartesian(p.xCoord(),p.yCoord(),c)

    def move (dx:Double,dy:Double):CPoint = 
      addColor(del.move(dx,dy),c)

    def add (q:CPoint):CPoint = 
      addColor(del.add(q),q.color())
    
    def rotate (t:Double):CPoint = 
      addColor(del.rotate(t),c)


    // methods specific to colored points:

    def color ():Color = c

    def updateColor (nc:Color):CPoint = 
      new CartesianCPoint(xpos,ypos,nc)
    


    // BRIDGE METHODS (some by delegation):

    def distance (q:Point):Double = del.distance(q)
    def add (q:Point):Point = del.add(q)

    def isEqual (q:Point):Boolean = q match {
      case cq:CPoint => this.isEqual(cq)
      case _ => false
    }
    

    // CANONICAL METHODS

    override def toString ():String = 
      "cartesian(" + xpos + "," + ypos + "," + c + ")"

    override def equals (other : Any):Boolean = 
      other match {
	case that : CPoint => this.isEqual(that)
	case _ => false
      }

    override def hashCode ():Int = 
      41 * (
	41 * (
	  41 + xpos.hashCode()
	) + ypos.hashCode()
      ) + c.hashCode()
  }
}



abstract class CPoint extends Point {

  def xCoord ():Double
  def yCoord ():Double
  def angleWithXAxis ():Double
  def distanceFromOrigin ():Double
  def color ():Color
  def updateColor (nc:Color):CPoint

  def distance (q:CPoint):Double
  def move (dx:Double,dy:Double):CPoint
  def add (q:CPoint):CPoint
  def rotate (theta:Double):CPoint
  def isEqual (q:CPoint):Boolean
  def isOrigin ():Boolean

  // Bridge methods
  def distance (q:Point):Double  
  def add (q:Point):Point        
  def isEqual (p:Point):Boolean
}
